@@ -184,6 +184,10 @@ static const luaL_Reg luv_functions[] = {
   {"pipe_chmod", luv_pipe_chmod},
 #endif
   {"pipe_connect", luv_pipe_connect},
+#if LUV_UV_VERSION_GEQ(1, 46, 0)
+  {"pipe_bind2", luv_pipe_bind2},
+  {"pipe_connect2", luv_pipe_connect2},
+#endif
   {"pipe_getsockname", luv_pipe_getsockname},
 #if LUV_UV_VERSION_GEQ(1, 3, 0)
   {"pipe_getpeername", luv_pipe_getpeername},
@@ -223,6 +227,9 @@ static const luaL_Reg luv_functions[] = {
   {"udp_set_ttl", luv_udp_set_ttl},
   {"udp_send", luv_udp_send},
   {"udp_try_send", luv_udp_try_send},
+#if LUV_UV_VERSION_GEQ(1, 50, 0)
+  {"udp_try_send2", luv_udp_try_send2},
+#endif
   {"udp_recv_start", luv_udp_recv_start},
   {"udp_recv_stop", luv_udp_recv_stop},
 #if LUV_UV_VERSION_GEQ(1, 27, 0)
@@ -326,6 +333,9 @@ static const luaL_Reg luv_functions[] = {
   {"setgid", luv_setgid},
 #endif
   {"getrusage", luv_getrusage},
+#if LUV_UV_VERSION_GEQ(1, 50, 0)
+  {"getrusage_thread", luv_getrusage_thread},
+#endif
   {"guess_handle", luv_guess_handle},
   {"hrtime", luv_hrtime},
   {"interface_addresses", luv_interface_addresses},
@@ -388,6 +398,22 @@ static const luaL_Reg luv_functions[] = {
   {"thread_setaffinity", luv_thread_setaffinity},
   {"thread_getcpu", luv_thread_getcpu},
 #endif
+#if LUV_UV_VERSION_GEQ(1, 48, 0)
+  {"thread_getpriority", luv_thread_getpriority},
+  {"thread_setpriority", luv_thread_setpriority},
+#endif
+#if LUV_UV_VERSION_GEQ(1, 50, 0)
+  {"thread_detach", luv_thread_detach},
+  {"thread_getname", luv_thread_getname},
+  {"thread_setname", luv_thread_setname},
+#endif
+
+#if LUV_UV_VERSION_GEQ(1, 49, 0)
+  {"utf16_length_as_wtf8", luv_utf16_length_as_wtf8},
+  {"utf16_to_wtf8", luv_utf16_to_wtf8},
+  {"wtf8_length_as_utf16", luv_wtf8_length_as_utf16},
+  {"wtf8_to_utf16", luv_wtf8_to_utf16},
+#endif
 
   // work.c
   {"new_work", luv_new_work},
@@ -485,6 +511,10 @@ static const luaL_Reg luv_pipe_methods[] = {
   {"chmod", luv_pipe_chmod},
 #endif
   {"connect", luv_pipe_connect},
+#if LUV_UV_VERSION_GEQ(1, 46, 0)
+  {"bind2", luv_pipe_bind2},
+  {"connect2", luv_pipe_connect2},
+#endif
   {"getsockname", luv_pipe_getsockname},
 #if LUV_UV_VERSION_GEQ(1, 3, 0)
   {"getpeername", luv_pipe_getpeername},
@@ -566,6 +596,9 @@ static const luaL_Reg luv_udp_methods[] = {
   {"set_ttl", luv_udp_set_ttl},
   {"send", luv_udp_send},
   {"try_send", luv_udp_try_send},
+#if LUV_UV_VERSION_GEQ(1, 50, 0)
+  {"try_send2", luv_udp_try_send2},
+#endif
   {"recv_start", luv_udp_recv_start},
   {"recv_stop", luv_udp_recv_stop},
 #if LUV_UV_VERSION_GEQ(1, 27, 0)
@@ -662,9 +695,8 @@ static void luv_req_init(lua_State* L) {
   lua_setfield(L, -2, "__index");
   lua_pop(L, 1);
 
-  // Only used for things that need to be garbage collected
-  // (e.g. the req when using uv_fs_scandir)
-  luaL_newmetatable(L, "uv_fs");
+  // Only used for luv_fs_scandir_t
+  luaL_newmetatable(L, "uv_fs_scandir");
   lua_pushcfunction(L, luv_req_tostring);
   lua_setfield(L, -2, "__tostring");
   luaL_newlib(L, luv_req_methods);
@@ -697,8 +729,11 @@ LUALIB_API int luv_cfpcall(lua_State* L, int nargs, int nresult, int flags) {
   case LUA_OK:
     break;
   case LUA_ERRMEM:
-    if ((flags & LUVF_CALLBACK_NOERRMSG) == 0)
-      fprintf(stderr, "System Error: %s\n", lua_tostring(L, -1));
+    if ((flags & LUVF_CALLBACK_NOERRMSG) == 0) {
+      fprintf(stderr, "System Error: %s\n",
+              luaL_tolstring(L, lua_absindex(L, -1), NULL));
+      lua_pop(L, 1); // Remove error string pushed by luaL_tolstring()
+    }
     if ((flags & LUVF_CALLBACK_NOEXIT) == 0)
       exit(-1);
     lua_pop(L, 1);
@@ -707,8 +742,11 @@ LUALIB_API int luv_cfpcall(lua_State* L, int nargs, int nresult, int flags) {
   case LUA_ERRRUN:
   case LUA_ERRERR:
   default:
-    if ((flags & LUVF_CALLBACK_NOERRMSG) == 0)
-      fprintf(stderr, "Uncaught Error: %s\n", lua_tostring(L, -1));
+    if ((flags & LUVF_CALLBACK_NOERRMSG) == 0) {
+      fprintf(stderr, "Uncaught Error: %s\n",
+              luaL_tolstring(L, lua_absindex(L, -1), NULL));
+      lua_pop(L, 1); // Remove error string pushed by luaL_tolstring()
+    }
     if ((flags & LUVF_CALLBACK_NOEXIT) == 0)
       exit(-1);
     lua_pop(L, 1);
@@ -750,6 +788,9 @@ LUALIB_API luv_ctx_t* luv_context(lua_State* L) {
     ctx = (luv_ctx_t*)lua_newuserdata(L, sizeof(*ctx));
     memset(ctx, 0, sizeof(*ctx));
     lua_rawset(L, LUA_REGISTRYINDEX);
+    // create table to contain internal handle
+    lua_newtable(L);
+    ctx->ht_ref = luaL_ref(L, LUA_REGISTRYINDEX);
   } else {
     ctx = (luv_ctx_t*)lua_touserdata(L, -1);
   }
@@ -815,6 +856,15 @@ static int loop_gc(lua_State *L) {
 }
 
 LUALIB_API int luaopen_luv (lua_State* L) {
+#ifdef LUA_RIDX_MAINTHREAD
+  // Lua 5.2+ - resolve the main thread of the current Lua state, even if
+  // we were loaded from a different thread (which may become suspended/dead).
+  lua_geti(L, LUA_REGISTRYINDEX, LUA_RIDX_MAINTHREAD);
+  lua_State* ctxL = lua_tothread(L, -1);
+  lua_pop(L, 1);
+#else
+  lua_State* ctxL = L;
+#endif
   luv_ctx_t* ctx = luv_context(L);
 
   luaL_newlib(L, luv_functions);
@@ -842,19 +892,13 @@ LUALIB_API int luaopen_luv (lua_State* L) {
     lua_rawset(L, -3);
 
     ctx->loop = loop;
-    ctx->L = L;
+    ctx->L = ctxL;
     ctx->mode = -1;
 
     ret = uv_loop_init(loop);
     if (ret < 0) {
       return luaL_error(L, "%s: %s\n", uv_err_name(ret), uv_strerror(ret));
     }
-
-    /* do cleanup in main thread */
-    lua_getglobal(L, "_THREAD");
-    if (lua_isnil(L, -1))
-      atexit(luv_work_cleanup);
-    lua_pop(L, 1);
   }
   // pcall is NULL, luv use default callback routine
   if (ctx->cb_pcall==NULL) {
